@@ -165,10 +165,9 @@ static void version_barf(void)
 #ifndef VCSID
 # define VCSID "<unknown>"
 #endif
-	printf("portage-utils-%s: compiled on %s\n"
-	       "vcs id: %s\n"
+	printf("portage-utils-%s: %s\n"
 	       "%s written for Gentoo by <solar and vapier @ gentoo.org>\n",
-	       VERSION, __DATE__, VCSID, argv0);
+	       VERSION, VCSID, argv0);
 	exit(EXIT_SUCCESS);
 }
 
@@ -1240,24 +1239,6 @@ char *atom_to_pvr(depend_atom *atom) {
 	return (atom->PR_int == 0 ? atom->P : atom->PVR );
 }
 
-/* TODO: Delete this in favor of libq/vdb.c API. */
-static char *grab_vdb_item(const char *item, const char *CATEGORY, const char *PF)
-{
-	static char *buf;
-	static size_t buf_len;
-
-	if (buf == NULL) {
-		buf_len = _Q_PATH_MAX;
-		buf = xmalloc(buf_len);
-	}
-
-	snprintf(buf, buf_len, "%s%s/%s/%s/%s", portroot, portvdb, CATEGORY, PF, item);
-	eat_file(buf, &buf, &buf_len);
-	rmspace(buf);
-
-	return buf;
-}
-
 /* TODO: Merge this into libq/vdb.c somehow. */
 _q_static queue *get_vdb_atoms(int fullcpv)
 {
@@ -1308,7 +1289,7 @@ _q_static queue *get_vdb_atoms(int fullcpv)
 				snprintf(buf, sizeof(buf), "%s/%s", atom->CATEGORY, atom->PN);
 			}
 			atom_implode(atom);
-			cpf = add_set(buf, slot, cpf);
+			cpf = add_set(buf, cpf);
 		}
 		scandir_free(pf, dfd);
 	}
