@@ -35,7 +35,7 @@
 
 int old_repo = 0;
 
-#define QMERGE_FLAGS "fFsKUpuyO5" COMMON_FLAGS
+#define QMERGE_FLAGS "fFsKUpuyO" COMMON_FLAGS
 static struct option const qmerge_long_opts[] = {
 	{"fetch",   no_argument, NULL, 'f'},
 	{"force",   no_argument, NULL, 'F'},
@@ -46,7 +46,6 @@ static struct option const qmerge_long_opts[] = {
 	{"update",  no_argument, NULL, 'u'},
 	{"yes",     no_argument, NULL, 'y'},
 	{"nodeps",  no_argument, NULL, 'O'},
-	{"nomd5",   no_argument, NULL, '5'},
 	{"debug",   no_argument, NULL, 128},
 	COMMON_LONG_OPTS
 };
@@ -60,7 +59,6 @@ static const char * const qmerge_opts_help[] = {
 	"Update only",
 	"Don't prompt before overwriting",
 	"Don't merge dependencies",
-	"Don't verify MD5 digest of files",
 	"Run shell funcs with `set -x`",
 	COMMON_OPTS_HELP
 };
@@ -72,7 +70,6 @@ char install = 0;
 char uninstall = 0;
 char force_download = 0;
 char follow_rdepends = 1;
-char nomd5 = 0;
 char qmerge_strict = 0;
 char update_only = 0;
 bool debug = false;
@@ -95,7 +92,7 @@ struct pkg_t {
 	size_t SIZE;
 	char USE[BUFSIZ];
 	char REPO[64];
-} Pkg;
+};
 
 struct llist_char_t {
 	char *data;
@@ -106,7 +103,7 @@ typedef struct llist_char_t llist_char;
 
 _q_static void pkg_fetch(int, const depend_atom *, const struct pkg_t *);
 _q_static void pkg_merge(int, const depend_atom *, const struct pkg_t *);
-_q_static int pkg_unmerge(const char *, const char *, queue *);
+_q_static int pkg_unmerge(q_vdb_pkg_ctx *, queue *);
 _q_static struct pkg_t *grab_binpkg_info(const char *);
 _q_static char *find_binpkg(const char *);
 
@@ -417,7 +414,7 @@ qprint_tree_node(int level, const depend_atom *atom, const struct pkg_t *pkg)
 }
 
 _q_static void
-pkg_run_func(const char *vdb_path, const char *phases, const char *func, const char *D, const char *T)
+pkg_run_func_at(int dirfd, const char *vdb_path, const char *phases, const char *func, const char *D, const char *T)
 {
 	const char *phase;
 	char *script;
@@ -444,7 +441,8 @@ pkg_run_func(const char *vdb_path, const char *phases, const char *func, const c
 		/* best_version() */
 		"use() { useq \"$@\"; }\n"
 		"usex() { useq \"$1\" && echo \"${2-yes}$4\" || echo \"${3-no}$5\"; }\n"
-		"useq() { hasq \"$1\" $USE; }\n"
+		"useq() { hasq \"$1\" ${USE}; }\n"
+		"usev() { hasv \"$1\" ${USE}; }\n"
 		"has() { hasq \"$@\"; }\n"
 		"hasq() { local h=$1; shift; case \" $* \" in *\" $h \"*) return 0;; *) return 1;; esac; }\n"
 		"hasv() { hasq \"$@\" && echo \"$1\"; }\n"
@@ -454,8 +452,11 @@ pkg_run_func(const char *vdb_path, const char *phases, const char *func, const c
 		"eqawarn() { elog \"QA: \"\"$@\"; }\n"
 		"eerror() { elog \"$@\"; }\n"
 		"die() { eerror \"$@\"; exit 1; }\n"
+		/* TODO: This should suppress `die` */
+		"nonfatal() { \"$@\"; }\n"
 		"ebegin() { printf ' * %%b ...' \"$*\"; }\n"
 		"eend() { local r=${1:-$?}; [ $# -gt 0 ] && shift; [ $r -eq 0 ] && echo ' [ ok ]' || echo \" $* \"'[ !! ]'; return $r; }\n"
+		/* TODO: This should be fatal upon error */
 		"emake() { ${MAKE:-make} ${MAKEOPTS} \"$@\"; }\n"
 		/* Unpack the env if need be */
 		"[ -e '%1$s/environment' ] || { bzip2 -dc '%1$s/environment.bz2' > '%1$s/environment' || exit 1; }\n"
@@ -466,9 +467,9 @@ pkg_run_func(const char *vdb_path, const char *phases, const char *func, const c
 		"MERGE_TYPE=binary\n"
 		"ROOT='%4$s'\n"
 		"EROOT=\"${EPREFIX%%/}/${ROOT#/}\"\n"
-		"D='%5$s'\n"
+		"D=\"%5$s\"\n"
 		"ED=\"${EPREFIX%%/}/${D#/}\"\n"
-		"T='%6$s'\n"
+		"T=\"%6$s\"\n"
 		/* Finally run the func */
 		"%7$s%2$s\n"
 		/* Ignore func return values (not exit values) */
@@ -480,9 +481,10 @@ pkg_run_func(const char *vdb_path, const char *phases, const char *func, const c
 		/*5*/ D,
 		/*6*/ T,
 		/*7*/ debug ? "set -x;" : "");
-	xsystembash(script);
+	xsystembash(script, dirfd);
 	free(script);
 }
+#define pkg_run_func(...) pkg_run_func_at(AT_FDCWD, __VA_ARGS__)
 
 /* Copy one tree (the single package) to another tree (ROOT) */
 _q_static int
@@ -562,7 +564,7 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 #if 0		/* We filter out "dir" as it's generally unnecessary cruft */
 			/* syntax: dir dirname */
 			fprintf(contents, "dir %s\n", cpath);
-			*objs = add_set(cpath, "", *objs);
+			*objs = add_set(cpath, *objs);
 			qprintf("%s>>>%s %s%s%s/\n", GREEN, NORM, DKBLUE, cpath, NORM);
 #endif
 
@@ -609,7 +611,7 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 				dname = name;
 				qprintf("%s>>>%s %s\n", GREEN, NORM, cpath);
 			}
-			*objs = add_set(cpath, "", *objs);
+			*objs = add_set(cpath, *objs);
 
 			/* First try fast path -- src/dst are same device */
 			if (renameat(subfd_src, dname, subfd_dst, name) == 0)
@@ -685,7 +687,7 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 			/* syntax: sym src -> dst mtime */
 			fprintf(contents, "sym %s -> %s %"PRIu64"u\n", cpath, sym, (uint64_t)st.st_mtime);
 			qprintf("%s>>>%s %s%s -> %s%s\n", GREEN, NORM, CYAN, cpath, sym, NORM);
-			*objs = add_set(cpath, "", *objs);
+			*objs = add_set(cpath, *objs);
 
 			/* Make it in the dest tree */
 			if (symlinkat(sym, subfd_dst, name)) {
@@ -753,6 +755,8 @@ _q_static void
 pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 {
 	queue *objs;
+	q_vdb_ctx *vdb_ctx;
+	q_vdb_cat_ctx *cat_ctx;
 	FILE *fp, *contents;
 	static char *phases;
 	static size_t phases_len;
@@ -860,6 +864,20 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	if (pretend)
 		return;
 
+	/* Get a handle on the main vdb repo */
+	vdb_ctx = q_vdb_open();
+	if (!vdb_ctx)
+		return;
+	cat_ctx = q_vdb_open_cat(vdb_ctx, pkg->CATEGORY);
+	if (!cat_ctx) {
+		if (errno != ENOENT)
+			return;
+		mkdirat(vdb_ctx->vdb_fd, pkg->CATEGORY, 0755);
+		cat_ctx = q_vdb_open_cat(vdb_ctx, pkg->CATEGORY);
+		if (!cat_ctx)
+			return;
+	}
+
 	/* Set up our temp dir to unpack this stuff */
 	xasprintf(&p, "%s/qmerge/%s/%s", port_tmpdir, pkg->CATEGORY, pkg->PF);
 	mkdir_p(p, 0755);
@@ -937,58 +955,44 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 
 	/* FIXME */ /* move unmerging to around here ? */
 	/* check for an already installed pkg */
-	snprintf(buf, sizeof(buf), "%s/%s", atom->CATEGORY, pkg->PF);
 
 	/* Unmerge any stray pieces from the older version which we didn't replace */
-	p = best_version(atom->CATEGORY, atom->PN);
-	if (*p) {
-		/* XXX: Should see about merging with unmerge_packages() */
-		makeargv(p, &ARGC, &ARGV);
-		for (i = 1; i < ARGC; i++) {
-			int ret, u;
-			const char *pn;
-			char *pf;
-			char *slot = NULL;
+	/* XXX: Should see about merging with unmerge_packages() */
+	while (1) {
+		int ret;
+		q_vdb_pkg_ctx *pkg_ctx;
+		depend_atom *old_atom;
 
-			pf = ARGV[i];
-			switch ((ret = atom_compare_str(buf, pf))) {
-				case ERROR:
-				case NOT_EQUAL:
-					continue;
-				case NEWER:
-				case OLDER:
-				case EQUAL:
-					u = 1;
-					pn = basename(pf);
-					slot = grab_vdb_item("SLOT", atom->CATEGORY, pn);
-					if (pkg->SLOT[0] && slot) {
-						if (strcmp(pkg->SLOT, slot) != 0)
-							u = 0;
-					}
-					/* We need to really set this unmerge pending after we look at contents of the new pkg */
-					if (u)
-						break;
-					continue;
-				default:
-					warn("no idea how we reached here.");
-					continue;
-			}
+		pkg_ctx = q_vdb_next_pkg(cat_ctx);
+		if (!pkg_ctx)
+			break;
 
-			qprintf("%s+++%s %s %s %s\n", GREEN, NORM, buf, booga[ret], pf);
-
-			pkg_unmerge(atom->CATEGORY, pn, objs);
+		old_atom = atom_explode(pkg_ctx->name);
+		/* This cast sucks, but we know for now the field isn't modified */
+		old_atom->CATEGORY = (char *)cat_ctx->name;
+		ret = atom_compare(atom, old_atom);
+		atom_implode(old_atom);
+		switch (ret) {
+			case NEWER:
+			case OLDER:
+			case EQUAL:
+				/* We need to really set this unmerge pending after we look at contents of the new pkg */
+				break;
+			default:
+				warn("no idea how we reached here.");
+			case ERROR:
+			case NOT_EQUAL:
+				continue;
 		}
-		freeargv(ARGC, ARGV);
+
+		qprintf("%s+++%s %s/%s %s %s/%s\n", GREEN, NORM, atom->CATEGORY, pkg->PF,
+			booga[ret], cat_ctx->name, pkg_ctx->name);
+
+		pkg_unmerge(pkg_ctx, objs);
 	}
 
 	/* Clean up the package state */
-	while (objs) {
-		queue *q = objs;
-		objs = q->next;
-		free(q->name);
-		free(q->item);
-		free(q);
-	}
+	free_sets(objs);
 	free(D);
 	free(T);
 
@@ -1018,68 +1022,45 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 }
 
 _q_static int
-pkg_unmerge(const char *cat, const char *pkgname, queue *keep)
+pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep)
 {
+	q_vdb_cat_ctx *cat_ctx = pkg_ctx->cat_ctx;
+	const char *cat = cat_ctx->name;
+	const char *pkgname = pkg_ctx->name;
 	size_t buflen;
 	static char *phases;
 	static size_t phases_len;
-	char *buf, *vdb_path, *T;
+	const char *T;
+	char *buf;
 	FILE *fp;
-	int ret, fd, vdb_fd, portroot_fd;
+	int ret, portroot_fd;
 	int cp_argc, cpm_argc;
 	char **cp_argv, **cpm_argv;
 	llist_char *dirs = NULL;
 
 	ret = 1;
-	buf = NULL;
-	vdb_path = NULL;
-	vdb_fd = portroot_fd = fd = -1;
+	buf = phases = NULL;
+	T = "${PWD}/temp";
 
-	if ((strchr(pkgname, ' ') != NULL) || (strchr(cat, ' ') != NULL)) {
-		qfprintf(stderr, "%s!!!%s '%s' '%s' (ambiguous name) specify fully-qualified pkgs\n", RED, NORM, cat, pkgname);
-		qfprintf(stderr, "%s!!!%s %s/%s (ambiguous name) specify fully-qualified pkgs\n", RED, NORM, cat, pkgname);
-		/* qfprintf(stderr, "%s!!!%s %s %s (ambiguous name) specify fully-qualified pkgs\n", RED, NORM, pkgname); */
-		return 1;
-	}
 	printf("%s<<<%s %s%s%s/%s%s%s\n", YELLOW, NORM, WHITE, cat, NORM, CYAN, pkgname, NORM);
 
 	if (pretend == 100)
 		return 0;
 
-	/* Get a handle to the root to play with */
-	portroot_fd = open(portroot, O_RDONLY|O_CLOEXEC|O_PATH);
-	if (portroot_fd == -1) {
-		warnp("unable to read %s", portroot);
-		goto done;
-	}
-
-	/* Get a handle on the vdb path which we'll use everywhere else */
-	/* Note: This vdb_path must be absolute since we use it in pkg_run_func() */
-	xasprintf(&vdb_path, "%s%s/%s/%s/", portroot, portvdb, cat, pkgname);
-	xasprintf(&T, "%stemp", vdb_path);
-	vdb_fd = openat(portroot_fd, vdb_path, O_RDONLY|O_CLOEXEC|O_PATH);
-	if (vdb_fd == -1) {
-		warnp("unable to read %s", vdb_path);
-		goto done;
-	}
-
-	/* First execute the pkg_prerm step */
-	if (!pretend) {
-		eat_file_at(vdb_fd, "DEFINED_PHASES", &phases, &phases_len);
-		mkdir_p(T, 0755);
-		pkg_run_func(vdb_path, phases, "pkg_prerm", T, T);
-	}
-
-	/* Now start removing all the installed files */
-	fd = openat(vdb_fd, "CONTENTS", O_RDONLY | O_CLOEXEC);
-	if (fd == -1) {
-		warnp("unable to read %s", "CONTENTS");
-		goto done;
-	}
-	fp = fdopen(fd, "r");
+	/* First get a handle on the things to clean up */
+	fp = q_vdb_pkg_fopenat_ro(pkg_ctx, "CONTENTS");
 	if (fp == NULL)
 		goto done;
 
+	portroot_fd = cat_ctx->ctx->portroot_fd;
+
+	/* Then execute the pkg_prerm step */
+	if (!pretend) {
+		q_vdb_pkg_eat(pkg_ctx, "DEFINED_PHASES", &phases, &phases_len);
+		mkdirat(pkg_ctx->fd, "temp", 0755);
+		pkg_run_func_at(pkg_ctx->fd, ".", phases, "pkg_prerm", T, T);
+	}
+
 	/* XXX: be nice to pull this out of the current func
 	 *      so we don't keep reparsing the same env var
 	 *      when unmerging multiple packages.
@@ -1181,7 +1162,6 @@ pkg_unmerge(const char *cat, const char *pkgname, queue *keep)
 	}
 
 	fclose(fp);
-	fd = -1;
 
 	/* Then remove all dirs in reverse order */
 	while (dirs != NULL) {
@@ -1193,9 +1173,9 @@ pkg_unmerge(const char *cat, const char *pkgname, queue *keep)
 		qprintf("%s%s%s %s%s%s/\n", rm ? YELLOW : GREEN, rm ? "---" : "<<<",
 			NORM, DKBLUE, dir, NORM);
 
+		dirs = dirs->next;
 		free(list->data);
 		free(list);
-		dirs = dirs->next;
 	}
 
 	freeargv(cp_argc, cp_argv);
@@ -1203,27 +1183,20 @@ pkg_unmerge(const char *cat, const char *pkgname, queue *keep)
 
 	if (!pretend) {
 		/* Then execute the pkg_postrm step */
-		pkg_run_func(vdb_path, phases, "pkg_postrm", T, T);
-		rm_rf(T);
+		pkg_run_func_at(pkg_ctx->fd, ".", phases, "pkg_postrm", T, T);
 
 		/* Finally delete the vdb entry */
-		rm_rf_at(portroot_fd, vdb_path);
+		rm_rf_at(pkg_ctx->fd, ".");
+		unlinkat(cat_ctx->fd, pkg_ctx->name, AT_REMOVEDIR);
 
-		/* And prune any empty vdb dirs */
-		rmdir_r_at(portroot_fd, vdb_path);
+		/* And prune the category if it's empty */
+		unlinkat(cat_ctx->ctx->vdb_fd, cat_ctx->name, AT_REMOVEDIR);
 	}
 
 	ret = 0;
  done:
-	if (fd != -1)
-		close(fd);
-	if (vdb_fd != -1)
-		close(vdb_fd);
-	if (portroot_fd != -1)
-		close(portroot_fd);
+	free(phases);
 	free(buf);
-	free(T);
-	free(vdb_path);
 
 	return ret;
 }
@@ -1237,49 +1210,6 @@ _q_static int unlink_empty(const char *buf)
 	return -1;
 }
 
-_q_static int match_pkg(queue *ll, const struct pkg_t *pkg)
-{
-	depend_atom *atom;
-	char buf[255], buf2[255];
-	int match = 0;
-
-	snprintf(buf, sizeof(buf), "%s/%s", pkg->CATEGORY, pkg->PF);
-	if ((atom = atom_explode(buf)) == NULL)
-		errf("%s/%s is not a valid atom", pkg->CATEGORY, pkg->PF);
-
-	/* verify this is the requested package */
-	if (strcmp(ll->name, buf) == 0)
-		match = 1;
-
-	if (strcmp(ll->name, pkg->PF) == 0)
-		match = 2;
-
-	snprintf(buf2, sizeof(buf2), "%s/%s", pkg->CATEGORY, atom->PN);
-
-	if (strcmp(ll->name, buf2) == 0)
-		match = 3;
-
-	if (strcmp(ll->name, atom->PN) == 0)
-		match = 4;
-
-	if (match)
-		goto match_done;
-
-	if (ll->item) {
-		depend_atom *subatom = atom_explode(ll->name);
-		if (subatom == NULL)
-			goto match_done;
-		if (strcmp(atom->PN, subatom->PN) == 0)
-			match = 1;
-		atom_implode(subatom);
-	}
-
-match_done:
-	atom_implode(atom);
-
-	return match;
-}
-
 _q_static int
 pkg_verify_checksums(char *fname, const struct pkg_t *pkg, const depend_atom *atom,
                      int strict, int display)
@@ -1287,9 +1217,6 @@ pkg_verify_checksums(char *fname, const struct pkg_t *pkg, const depend_atom *at
 	char *hash = NULL;
 	int ret = 0;
 
-	if (nomd5)
-		return ret;
-
 	if (pkg->MD5[0]) {
 		if ((hash = (char*) hash_file(fname, HASH_MD5)) == NULL) {
 			errf("hash is NULL for %s", fname);
@@ -1325,31 +1252,6 @@ pkg_verify_checksums(char *fname, const struct pkg_t *pkg, const depend_atom *at
 	return ret;
 }
 
-_q_static
-void pkg_process(queue *todo, const struct pkg_t *pkg)
-{
-	queue *ll;
-	depend_atom *atom;
-	char buf[255];
-
-	snprintf(buf, sizeof(buf), "%s/%s", pkg->CATEGORY, pkg->PF);
-	if ((atom = atom_explode(buf)) == NULL)
-		errf("%s/%s is not a valid atom", pkg->CATEGORY, pkg->PF);
-
-	ll = todo;
-	while (ll) {
-		if (ll->name[0] != '-' && match_pkg(ll, pkg)) {
-			/* fetch all requested packages */
-			pkg_fetch(0, atom, pkg);
-		}
-
-		ll = ll->next;
-	}
-
-	/* free the atom */
-	atom_implode(atom);
-}
-
 _q_static void
 pkg_fetch(int level, const depend_atom *atom, const struct pkg_t *pkg)
 {
@@ -1421,19 +1323,15 @@ pkg_fetch(int level, const depend_atom *atom, const struct pkg_t *pkg)
 }
 
 _q_static void
-print_Pkg(int full, struct pkg_t *pkg)
+print_Pkg(int full, const depend_atom *atom, const struct pkg_t *pkg)
 {
 	char *p = NULL;
 	char buf[512];
-	depend_atom *atom = NULL;
 
-	if (!pkg->CATEGORY[0]) errf("CATEGORY is NULL");
-	if (!pkg->PF[0]) errf("PF is NULL");
-
-	printf("%s%s/%s%s%s%s%s%s\n", BOLD, pkg->CATEGORY, BLUE, pkg->PF, NORM,
+	printf("%s%s/%s%s%s%s%s%s\n", BOLD, atom->CATEGORY, BLUE, pkg->PF, NORM,
 		!quiet ? " [" : "",
 		!quiet ? make_human_readable_str(pkg->SIZE, 1, KILOBYTE) : "",
-		!quiet ? "KiB]" : "");
+		!quiet ? " KiB]" : "");
 
 	if (full == 0)
 		return;
@@ -1458,10 +1356,6 @@ print_Pkg(int full, struct pkg_t *pkg)
 		if (strcmp(pkg->REPO, "gentoo") != 0)
 			printf(" %sRepo%s:%s %s\n", DKGREEN, YELLOW, NORM, pkg->REPO);
 
-	snprintf(buf, sizeof(buf), "%s/%s", pkg->CATEGORY, pkg->PF);
-	atom = atom_explode(buf);
-	if ((atom = atom_explode(buf)) == NULL)
-		return;
 	if ((p = best_version(pkg->CATEGORY, atom->PN)) != NULL) {
 		if (*p) {
 			int ret;
@@ -1476,44 +1370,26 @@ print_Pkg(int full, struct pkg_t *pkg)
 			printf(" %sInstalled%s:%s %s%s%s\n", DKGREEN, YELLOW, NORM, icolor, p, NORM);
 		}
 	}
-	atom_implode(atom);
+}
+
+_q_static int
+qmerge_unmerge_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
+{
+	queue *todo = priv;
+
+	while (todo) {
+		if (qlist_match(pkg_ctx, todo->name, NULL, true))
+			pkg_unmerge(pkg_ctx, NULL);
+		todo = todo->next;
+	}
+
+	return 0;
 }
 
 _q_static int
 unmerge_packages(queue *todo)
 {
-	depend_atom *atom;
-	char *p;
-	int i, argc;
-	char **argv;
-
-	while (todo) {
-		char buf[512];
-
-		if (todo->name[0] == '-')
-			goto next;
-
-		p = best_version(NULL, todo->name);
-		if (!*p)
-			goto next;
-
-		makeargv(p, &argc, &argv);
-		for (i = 1; i < argc; ++i) {
-			if ((atom = atom_explode(argv[i])) == NULL)
-				continue;
-			if (atom->CATEGORY) {
-				atom2str(atom, buf, sizeof(buf));
-				pkg_unmerge(atom->CATEGORY, buf, NULL);
-			}
-			atom_implode(atom);
-		}
-		freeargv(argc, argv);
-
- next:
-		todo = todo->next;
-	}
-
-	return 0;
+	return q_vdb_foreach_pkg(qmerge_unmerge_cb, todo, NULL);
 }
 
 _q_static FILE *
@@ -1575,6 +1451,7 @@ grab_binpkg_info(const char *name)
 	char *p;
 	depend_atom *atom;
 
+	struct pkg_t Pkg;
 	struct pkg_t *pkg = xzalloc(sizeof(struct pkg_t));
 	struct pkg_t *rpkg = xzalloc(sizeof(struct pkg_t));
 
@@ -1620,28 +1497,28 @@ grab_binpkg_info(const char *name)
 			}
 			continue;
 		}
+
 		if ((p = strchr(buf, '\n')) != NULL)
 			*p = 0;
-
 		if ((p = strchr(buf, ':')) == NULL)
 			continue;
-		if ((p = strchr(buf, ' ')) == NULL)
+		if (p[1] != ' ')
 			continue;
 		*p = 0;
-		++p;
+		p += 2;
 
 		if (*buf) {
 			/* we dont need all the info */
-			if (strcmp(buf, "RDEPEND:") == 0)
+			if (strcmp(buf, "RDEPEND") == 0)
 				strncpy(pkg->RDEPEND, p, sizeof(Pkg.RDEPEND));
-			if (strcmp(buf, "PF:") == 0)
+			if (strcmp(buf, "PF") == 0)
 				strncpy(pkg->PF, p, sizeof(Pkg.PF));
-			if (strcmp(buf, "CATEGORY:") == 0)
+			if (strcmp(buf, "CATEGORY") == 0)
 				strncpy(pkg->CATEGORY, p, sizeof(Pkg.CATEGORY));
-			if (strcmp(buf, "REPO:") == 0)
+			if (strcmp(buf, "REPO") == 0)
 				strncpy(pkg->REPO, p, sizeof(Pkg.REPO));
 
-			if (strcmp(buf, "CPV:") == 0) {
+			if (strcmp(buf, "CPV") == 0) {
 				if ((atom = atom_explode(p)) != NULL) {
 					snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
 					if (atom->PR_int)
@@ -1651,15 +1528,15 @@ grab_binpkg_info(const char *name)
 					atom_implode(atom);
 				}
 			}
-			if (strcmp(buf, "SLOT:") == 0)
+			if (strcmp(buf, "SLOT") == 0)
 				strncpy(pkg->SLOT, p, sizeof(Pkg.SLOT));
-			if (strcmp(buf, "USE:") == 0)
+			if (strcmp(buf, "USE") == 0)
 				strncpy(pkg->USE, p, sizeof(Pkg.USE));
 
 			/* checksums. We must have 1 or the other unless --*/
-			if (strcmp(buf, "MD5:") == 0)
+			if (strcmp(buf, "MD5") == 0)
 				strncpy(pkg->MD5, p, sizeof(Pkg.MD5));
-			if (strcmp(buf, "SHA1:") == 0)
+			if (strcmp(buf, "SHA1") == 0)
 				strncpy(pkg->SHA1, p, sizeof(Pkg.SHA1));
 		}
 	}
@@ -1674,6 +1551,7 @@ find_binpkg(const char *name)
 	FILE *fp;
 	char buf[BUFSIZ];
 	char *p;
+	struct pkg_t Pkg;
 	char PF[sizeof(Pkg.PF)];
 	char CATEGORY[sizeof(Pkg.CATEGORY)];
 
@@ -1716,18 +1594,18 @@ find_binpkg(const char *name)
 			}
 			continue;
 		}
+
 		if ((p = strchr(buf, '\n')) != NULL)
 			*p = 0;
-
 		if ((p = strchr(buf, ':')) == NULL)
 			continue;
-		if ((p = strchr(buf, ' ')) == NULL)
+		if (p[1] != ' ')
 			continue;
 		*p = 0;
-		++p;
+		p += 2;
 
 		if (*buf) {
-			if (strcmp(buf, "CPV:") == 0) {
+			if (strcmp(buf, "CPV") == 0) {
 				depend_atom *atom;
 				if ((atom = atom_explode(p)) != NULL) {
 					snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
@@ -1738,9 +1616,9 @@ find_binpkg(const char *name)
 					atom_implode(atom);
 				}
 			}
-			if (strcmp(buf, "PF:") == 0)
+			if (strcmp(buf, "PF") == 0)
 				strncpy(PF, p, sizeof(PF));
-			if (strcmp(buf, "CATEGORY:") == 0)
+			if (strcmp(buf, "CATEGORY") == 0)
 				strncpy(CATEGORY, p, sizeof(CATEGORY));
 		}
 	}
@@ -1754,86 +1632,119 @@ parse_packages(queue *todo)
 	FILE *fp;
 	size_t buflen;
 	char *buf, *p;
-	long lineno = 0;
+	struct pkg_t Pkg;
+	depend_atom *pkg_atom;
+	char repo[sizeof(Pkg.REPO)];
 
 	fp = open_binpkg_index();
 
-	memset(&Pkg, 0, sizeof(Pkg));
-
 	buf = NULL;
+	repo[0] = '\0';
+
+	/* First consume the header with the common data. */
 	while (getline(&buf, &buflen, fp) != -1) {
-		lineno++;
-		if (*buf == '\n') {
-			if ((strlen(Pkg.PF) > 0) && (strlen(Pkg.CATEGORY) > 0)) {
-				struct pkg_t *pkg = xmalloc(sizeof(*pkg));
-				*pkg = Pkg;
+		if (*buf == '\n')
+			break;
 
-				if (search_pkgs) {
-					if (todo) {
-						queue *ll = todo;
-						while (ll) {
-							if ((match_pkg(ll, pkg) > 0) || (strcmp(ll->name, pkg->CATEGORY) == 0))
-								print_Pkg(verbose, pkg);
-							ll = ll->next;
-						}
-					} else
-						print_Pkg(verbose, pkg);
-				} else
-					pkg_process(todo, pkg);
-
-				free(pkg);
-			}
-			memset(&Pkg, 0, sizeof(Pkg));
-			continue;
-		}
 		if ((p = strchr(buf, '\n')) != NULL)
 			*p = 0;
-
 		if ((p = strchr(buf, ':')) == NULL)
 			continue;
-		if ((p = strchr(buf, ' ')) == NULL)
+		if (p[1] != ' ')
 			continue;
 		*p = 0;
-		++p;
+		p += 2;
+
+		switch (*buf) {
+		case 'R':
+			if (!strcmp(buf, "REPO"))
+				strncpy(repo, p, sizeof(repo));
+			break;
+		}
+	}
+
+	pkg_atom = NULL;
+	memset(&Pkg, 0, sizeof(Pkg));
+	strcpy(Pkg.SLOT, "0");
+
+	/* Then walk all the package entries. */
+	while (getline(&buf, &buflen, fp) != -1) {
+		if (*buf == '\n') {
+			if (pkg_atom) {
+				if (search_pkgs && !todo) {
+					print_Pkg(verbose, pkg_atom, &Pkg);
+				} else {
+					const queue *ll = todo;
+					depend_atom *todo_atom;
+					while (ll) {
+						todo_atom = atom_explode(ll->name);
+						pkg_atom->REPO = todo_atom->REPO ? Pkg.REPO : NULL;
+						pkg_atom->SLOT = todo_atom->SLOT ? Pkg.SLOT : NULL;
+						if (atom_compare(todo_atom, pkg_atom) == EQUAL) {
+							if (search_pkgs)
+								print_Pkg(verbose, pkg_atom, &Pkg);
+							else
+								pkg_fetch(0, pkg_atom, &Pkg);
+						}
+						atom_implode(todo_atom);
+						ll = ll->next;
+					}
+				}
+
+				atom_implode(pkg_atom);
+				pkg_atom = NULL;
+			}
+			memset(&Pkg, 0, sizeof(Pkg));
+			strcpy(Pkg.SLOT, "0");
+			strcpy(Pkg.REPO, repo);
+			continue;
+		}
+
+		if ((p = strchr(buf, '\n')) != NULL)
+			*p = 0;
+		if ((p = strchr(buf, ':')) == NULL)
+			continue;
+		if (p[1] != ' ')
+			continue;
+		*p = 0;
+		p += 2;
 
 		switch (*buf) {
 			case 'U':
-				if (strcmp(buf, "USE:") == 0) strncpy(Pkg.USE, p, sizeof(Pkg.USE));
+				if (strcmp(buf, "USE") == 0) strncpy(Pkg.USE, p, sizeof(Pkg.USE));
 				break;
 			case 'P':
-				if (strcmp(buf, "PF:") == 0) strncpy(Pkg.PF, p, sizeof(Pkg.PF));
+				if (strcmp(buf, "PF") == 0) strncpy(Pkg.PF, p, sizeof(Pkg.PF));
 				break;
 			case 'S':
-				if (strcmp(buf, "SIZE:") == 0) Pkg.SIZE = atol(p);
-				if (strcmp(buf, "SLOT:") == 0) strncpy(Pkg.SLOT, p, sizeof(Pkg.SLOT));
-				if (strcmp(buf, "SHA1:") == 0) strncpy(Pkg.SHA1, p, sizeof(Pkg.SHA1));
+				if (strcmp(buf, "SIZE") == 0) Pkg.SIZE = atol(p);
+				if (strcmp(buf, "SLOT") == 0) strncpy(Pkg.SLOT, p, sizeof(Pkg.SLOT));
+				if (strcmp(buf, "SHA1") == 0) strncpy(Pkg.SHA1, p, sizeof(Pkg.SHA1));
 				break;
 			case 'M':
-				if (strcmp(buf, "MD5:") == 0) strncpy(Pkg.MD5, p, sizeof(Pkg.MD5));
+				if (strcmp(buf, "MD5") == 0) strncpy(Pkg.MD5, p, sizeof(Pkg.MD5));
 				break;
 			case 'R':
-				if (strcmp(buf, "REPO:") == 0) strncpy(Pkg.REPO, p, sizeof(Pkg.REPO));
-				if (strcmp(buf, "RDEPEND:") == 0) strncpy(Pkg.RDEPEND, p, sizeof(Pkg.RDEPEND));
+				if (strcmp(buf, "REPO") == 0) strncpy(Pkg.REPO, p, sizeof(Pkg.REPO));
+				if (strcmp(buf, "RDEPEND") == 0) strncpy(Pkg.RDEPEND, p, sizeof(Pkg.RDEPEND));
 				break;
 			case 'L':
-				if (strcmp(buf, "LICENSE:") == 0) strncpy(Pkg.LICENSE, p, sizeof(Pkg.LICENSE));
+				if (strcmp(buf, "LICENSE") == 0) strncpy(Pkg.LICENSE, p, sizeof(Pkg.LICENSE));
 				break;
 			case 'C':
-				if (strcmp(buf, "CATEGORY:") == 0) strncpy(Pkg.CATEGORY, p, sizeof(Pkg.CATEGORY));
-				if (strcmp(buf, "CPV:") == 0) {
-					depend_atom *atom;
-					if ((atom = atom_explode(p)) != NULL) {
-						if (atom->PR_int)
-							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s-r%i", atom->PN, atom->PV, atom->PR_int);
+				if (strcmp(buf, "CATEGORY") == 0) strncpy(Pkg.CATEGORY, p, sizeof(Pkg.CATEGORY));
+				if (strcmp(buf, "CPV") == 0) {
+					if ((pkg_atom = atom_explode(p)) != NULL) {
+						if (pkg_atom->PR_int)
+							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s-r%i", pkg_atom->PN, pkg_atom->PV, pkg_atom->PR_int);
 						else
-							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s", atom->PN, atom->PV);
-						strncpy(Pkg.CATEGORY, atom->CATEGORY, sizeof(Pkg.CATEGORY));
-						atom_implode(atom);
+							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s", pkg_atom->PN, pkg_atom->PV);
+						strncpy(Pkg.CATEGORY, pkg_atom->CATEGORY, sizeof(Pkg.CATEGORY));
 					}
 				}
 				break;
 			case 'D':
-				if (strcmp(buf, "DESC:") == 0) strncpy(Pkg.DESC, p, sizeof(Pkg.DESC));
+				if (strcmp(buf, "DESC") == 0) strncpy(Pkg.DESC, p, sizeof(Pkg.DESC));
 				break;
 			default:
 				break;
@@ -1842,25 +1753,12 @@ parse_packages(queue *todo)
 
 	free(buf);
 	fclose(fp);
+	if (pkg_atom)
+		atom_implode(pkg_atom);
 
 	return EXIT_SUCCESS;
 }
 
-_q_static queue *
-qmerge_add_set_atom(char *satom, queue *set)
-{
-	char *p;
-	const char *slot;
-
-	if ((p = strchr(satom, ':')) != NULL) {
-		*p = 0;
-		slot = p + 1;
-	} else
-		slot = "0";
-
-	return add_set(satom, slot, set);
-}
-
 _q_static queue *
 qmerge_add_set_file(const char *dir, const char *file, queue *set)
 {
@@ -1882,7 +1780,7 @@ qmerge_add_set_file(const char *dir, const char *file, queue *set)
 	buf = NULL;
 	while (getline(&buf, &buflen, fp) != -1) {
 		rmspace(buf);
-		set = qmerge_add_set_atom(buf, set);
+		set = add_set(buf, set);
 	}
 	free(buf);
 
@@ -1904,7 +1802,7 @@ qmerge_add_set_system(void *data, char *buf)
 
 	s = buf;
 	if (*s == '*')
-		set = add_set(s + 1, "", set);
+		set = add_set(s + 1, set);
 	else if (s[0] == '-' && s[1] == '*') {
 		int ok;
 		set = del_set(s + 2, set, &ok);
@@ -1928,7 +1826,7 @@ qmerge_add_set(char *buf, queue *set)
 	else if (buf[0] == '@')
 		return qmerge_add_set_file("/etc/portage", buf+1, set);
 	else
-		return qmerge_add_set_atom(buf, set);
+		return add_set(buf, set);
 }
 
 _q_static int
@@ -1952,7 +1850,7 @@ int qmerge_main(int argc, char **argv)
 		switch (i) {
 			case 'f': force_download = 1; break;
 			case 'F': force_download = 2; break;
-			case 's': search_pkgs = 1; break;
+			case 's': search_pkgs = 1; interactive = 0; break;
 			/* case 'i': case 'g': */
 			case 'K': install = 1; break;
 			case 'U': uninstall = 1; break;
@@ -1960,7 +1858,6 @@ int qmerge_main(int argc, char **argv)
 			case 'u': update_only = 1;
 			case 'y': interactive = 0; break;
 			case 'O': follow_rdepends = 0; break;
-			case '5': nomd5 = 1; break;
 			case 128: debug = true; break;
 			COMMON_GETOPTS_CASES(qmerge)
 		}
